'use strict';

Object.defineProperty(exports, "__esModule", {
  value: true
});

var _slicedToArray = function () { function sliceIterator(arr, i) { var _arr = []; var _n = true; var _d = false; var _e = undefined; try { for (var _i = arr[Symbol.iterator](), _s; !(_n = (_s = _i.next()).done); _n = true) { _arr.push(_s.value); if (i && _arr.length === i) break; } } catch (err) { _d = true; _e = err; } finally { try { if (!_n && _i["return"]) _i["return"](); } finally { if (_d) throw _e; } } return _arr; } return function (arr, i) { if (Array.isArray(arr)) { return arr; } else if (Symbol.iterator in Object(arr)) { return sliceIterator(arr, i); } else { throw new TypeError("Invalid attempt to destructure non-iterable instance"); } }; }();

exports.default = mapper;

var _splitString = require('split-string');

var _splitString2 = _interopRequireDefault(_splitString);

var _isPrimitive = require('is-primitive');

var _isPrimitive2 = _interopRequireDefault(_isPrimitive);

var _compose = require('./compose');

var _compose2 = _interopRequireDefault(_compose);

var _is = require('./is');

var _is2 = _interopRequireDefault(_is);

var _utils = require('./utils');

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

function _toConsumableArray(arr) { if (Array.isArray(arr)) { for (var i = 0, arr2 = Array(arr.length); i < arr.length; i++) { arr2[i] = arr[i]; } return arr2; } else { return Array.from(arr); } }

var separator = '.';
var replaceReg = /^\$:/;
var formatReg = /^\$:.+/;
var sameThing = function sameThing(value) {
  return value;
};

var formaterCheck = function formaterCheck(formater) {
  return _is2.default.String(formater) && formatReg.test(formater);
};

var getValueFinder = function getValueFinder(source) {
  return function (keys) {
    var transform = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : sameThing;

    return keys.reduce(function (value, key) {
      try {
        return transform(value[key]);
      } catch (err) {
        (0, _utils.log)('error', 'Hope: A error has occurred when read the ' + JSON.stringify(keys) + '.', err);
        return value;
      }
    }, source);
  };
};

var getFormater = function getFormater(formater) {
  // formater like [ format-string like '$:a.b.c', transform, transform1, ... ]
  if (_is2.default.Array(formater)) {
    var format = formater[0];
    var funcs = formater.slice(1, formater.length);
    (0, _utils.invariant)(formater.length >= 2 && !funcs.some(function (func) {
      return !_is2.default.Function(func);
    }) && _is2.default.String(format) && formaterCheck(format), "Hope: Expecting formater is a array like [ '$:..', ...Functions ].");
    return [format, _compose2.default.apply(undefined, _toConsumableArray(funcs))];
  }

  // string like '$:a.b.c' or values
  return [formater];
};

/**
 * @param {Object} options {
 *   spliter {RegExp}
 *   map {Object}
 * }
 * @return {Function}
 */
function mapper() {
  var options = arguments.length > 0 && arguments[0] !== undefined ? arguments[0] : {};

  var format = options.format;
  var sep = options.sep || separator;

  (0, _utils.invariant)(_is2.default.PlainObject(format), 'Hope: Expecting format of mapper is a plain-object, but got ' + (0, _is.kindOf)(format));

  var getKeys = function getKeys(str) {
    // 'as.d\\.asa.ds' => [ 'as', 'd\\.asa', 'ds' ]
    return (0, _splitString2.default)(str.replace(replaceReg, ''), { sep: sep, keepEscaping: true });
  };

  return function compiler(source) {
    (0, _utils.invariant)(!(0, _isPrimitive2.default)(source), 'Hope: Expecting source of mapper is a plain-object or array, but got ' + (0, _is.kindOf)(source));

    var finder = getValueFinder(source);

    /**
     * if the formater is a string like '$:..'
     * it will be replaced with the found value
     */
    function findValue(_format) {
      var next = {};
      Object.keys(_format).forEach(function (key) {
        var value = _format[key];
        if (_is2.default.PlainObject(value)) {
          next[key] = findValue(value);
        } else {
          var _getFormater = getFormater(value),
              _getFormater2 = _slicedToArray(_getFormater, 2),
              formater = _getFormater2[0],
              transform = _getFormater2[1];

          if (formaterCheck(formater)) {
            next[key] = finder(getKeys(formater), transform);
          } else {
            next[key] = formater;
          }
        }
      });
      return next;
    }

    return findValue(format);
  };
}

var isNode = false;

try {
  window.document;
} catch (e) {
  isNode = true;
}
if (!isNode) {
  window.mapper = mapper;
}
module.exports = exports['default'];