'use strict';

Object.defineProperty(exports, "__esModule", {
  value: true
});

var _omit = require('omit.js');

var _omit2 = _interopRequireDefault(_omit);

var _ajax = require('./ajax/ajax');

var _ajax2 = _interopRequireDefault(_ajax);

var _compose = require('./compose');

var _compose2 = _interopRequireDefault(_compose);

var _utils = require('./utils');

var _is = require('./is');

var _is2 = _interopRequireDefault(_is);

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

function _objectWithoutProperties(obj, keys) { var target = {}; for (var i in obj) { if (keys.indexOf(i) >= 0) continue; if (!Object.prototype.hasOwnProperty.call(obj, i)) continue; target[i] = obj[i]; } return target; }

var noop = function noop() {};

/**
 * options {
 *   interval = 1000
 * }
 */

var defaultOption = {
  interval: 1000,
  retryTimes: 10,
  timeout: 30 * 60 * 1000,
  getToken: noop
};

function getOption(_ref) {
  var interval = _ref.interval,
      timeout = _ref.timeout,
      retryTimes = _ref.retryTimes,
      getToken = _ref.getToken,
      other = _objectWithoutProperties(_ref, ['interval', 'timeout', 'retryTimes', 'getToken']);

  var options = Object.assign(other, defaultOption);

  if (!_is2.default.Number(interval) || !isFinite(interval)) {
    options.interval = 1000;
  } else {
    options.interval = interval;
  }

  if (_is2.default.Number(timeout) && isFinite(timeout)) {
    options.timeout = Math.max(0, timeout);
  }

  if (_is2.default.Number(retryTimes)) {
    options.retryTimes = retryTimes > 0 ? retryTimes : defaultOption.retryTimes;
  }

  if (_is2.default.Function(getToken)) {
    options.getToken = getToken;
  }

  return options;
}

function Polling(options) {

  var xhr = void 0;
  var timer = void 0;
  var started = void 0;
  var interval = 0;
  var retryTimes = options.retryTimes;
  var callbacks = noop;
  var returnMethod = {};

  function alwaysHandle(err, results) {
    if (!err) {
      interval = 0;
      retryTimes = options.retryTimes;
      var token = options.getToken(results[0]);
      callbacks({ token: token, result: results[0] });
    } else {
      interval = Math.min(interval + options.interval, 6000);
      retryTimes -= 1;
    }
  }

  function send() {
    timer = null;
    var ajaxOptions = (0, _omit2.default)(options, ['interval', 'retryTimes', 'getToken']);

    var url = ajaxOptions.url,
        otherOptions = _objectWithoutProperties(ajaxOptions, ['url']);

    xhr = (0, _ajax2.default)(url, otherOptions).always(alwaysHandle).always(function (err) {
      if (started) {
        xhr = null;
        if (!err || err && retryTimes > 0) {
          timer = setTimeout(send, interval);
          (0, _utils.log)('log', 'Polling will try to conenct after ' + interval / 1000 + ' second.');
        } else if (err && retryTimes === 0) {
          (0, _utils.log)('warn', 'Polling have tried to conenct ' + options.retryTimes + ' times.');
        }
      }
    });
    return xhr;
  }

  function start() {
    if (!started) {
      started = true;
      send();
    }
    return returnMethod;
  }

  function stop() {
    started = false;
    if (xhr) {
      xhr.abort();
      xhr = null;
    }
    if (timer) {
      clearTimeout(timer);
      timer = null;
    }
    return returnMethod;
  }

  function maybe(token_, callback) {
    return function (set) {
      var token = set.token,
          result = set.result;

      if (_is2.default.Function(token_)) {
        token_(result);
      } else if (_is2.default.Undefined(token) || token_ === token) {
        callback(result);
      }
      return set;
    };
  }

  var onResponse = (0, _compose2.default)(function (cb) {
    callbacks = callbacks !== noop ? (0, _compose2.default)(callbacks, cb) : cb;
    return returnMethod;
  }, maybe);

  Object.assign(returnMethod, { start: start, stop: stop, onResponse: onResponse });

  return returnMethod;
}

exports.default = (0, _compose2.default)(Polling, getOption);
module.exports = exports['default'];