'use strict';

Object.defineProperty(exports, "__esModule", {
  value: true
});

var _events = require('events');

var _events2 = _interopRequireDefault(_events);

var _is = require('./is');

var _is2 = _interopRequireDefault(_is);

var _utils = require('./utils');

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

function _toConsumableArray(arr) { if (Array.isArray(arr)) { for (var i = 0, arr2 = Array(arr.length); i < arr.length; i++) { arr2[i] = arr[i]; } return arr2; } else { return Array.from(arr); } }

/**
 * cache some values those use globally or will appear in future
 */

var collection = new Map();
var emitter = new _events2.default();
var set = collection.set.bind(collection);
var get = collection.get.bind(collection);
var getAll = function getAll() {
  return [].concat(_toConsumableArray(collection));
};
var wait = function wait(key, func, always) {

  (0, _utils.invariant)(_is2.default.Defined(key), 'Hope: key of store.wait should not be undefined/null');

  (0, _utils.invariant)(_is2.default.Function(func), 'Hope: Expecting func of store.wait is a function, but got ' + (0, _is.kindOf)(func));

  var value = get(key);
  function clear() {
    // eslint-disable-next-line
    emitter.removeListener(key, handle);
  }

  function handle() {
    func(get(key), clear);
  }

  if (always) {
    emitter.on(key, handle);
  } else if (value !== undefined) {
    func(value, clear);
  } else {
    emitter.once(key, handle);
  }

  return { key: key, clear: clear };
};

var watch = function watch(key, func) {
  return wait(key, func, true);
};

var dispense = function dispense(key, value) {
  if (key) {
    if (value !== undefined) {
      set(key, value);
    }
    emitter.emit(key);
  }
};

var clear = function clear() {
  collection = new Map();
};

emitter.setMaxListeners(1000);

exports.default = {
  set: set,
  get: get,
  getAll: getAll,
  wait: wait,
  watch: watch,
  dispense: dispense,
  clear: clear
};
module.exports = exports['default'];